using System;
using System.Collections;
using System.Drawing;
using System.Reflection;
using System.IO;

using org.zefer.pd4ml;
using org.zefer.html.parser;
using org.zefer.html.doc;
using org.zefer.pd4ml.pdf;

namespace PD4Test
{
    class Tokenizer
    {
        //Element list identified
        private System.Collections.ArrayList elements;
        //Source string to use
        private string source;
        //The tokenizer uses the default delimiter set: the space character, the tab character, the newline character, and the carriage-return character
        private string delimiters = " \t\n\r";      

        /// <summary>
        /// Initializes a new class instance with a specified string to process
        /// </summary>
        /// <param name="source">String to tokenize</param>
        public Tokenizer(string source)
        {           
            this.elements = new System.Collections.ArrayList();
            this.elements.AddRange(source.Split(this.delimiters.ToCharArray()));
            this.RemoveEmptyStrings();
            this.source = source;
        }

        /// <summary>
        /// Initializes a new class instance with a specified string to process
        /// and the specified token delimiters to use
        /// </summary>
        /// <param name="source">String to tokenize</param>
        /// <param name="delimiters">String containing the delimiters</param>
        public Tokenizer(string source, string delimiters)
        {
            this.elements = new System.Collections.ArrayList();
            this.delimiters = delimiters;
            this.elements.AddRange(source.Split(this.delimiters.ToCharArray()));
            this.RemoveEmptyStrings();
            this.source = source;
        }

        /// <summary>
        /// Current token count for the source string
        /// </summary>
        public int Count
        {
            get
            {
                return (this.elements.Count);
            }
        }

        /// <summary>
        /// Determines if there are more tokens to return from the source string
        /// </summary>
        /// <returns>True or false, depending if there are more tokens</returns>
        public bool HasMoreTokens()
        {
            return (this.elements.Count > 0);           
        }

        /// <summary>
        /// Returns the next token from the token list
        /// </summary>
        /// <returns>The string value of the token</returns>
        public string NextToken()
        {           
            string result;
            if (source == "") throw new System.Exception();
            else
            {
                this.elements = new System.Collections.ArrayList();
                this.elements.AddRange(this.source.Split(delimiters.ToCharArray()));
                RemoveEmptyStrings();       
                result = (string) this.elements[0];
                this.elements.RemoveAt(0);              
                this.source = this.source.Remove(this.source.IndexOf(result),result.Length);
                this.source = this.source.TrimStart(this.delimiters.ToCharArray());
                return result;                  
            }           
        }

        /// <summary>
        /// Returns the next token from the source string, using the provided
        /// token delimiters
        /// </summary>
        /// <param name="delimiters">String containing the delimiters to use</param>
        /// <returns>The string value of the token</returns>
        public string NextToken(string delimiters)
        {
            this.delimiters = delimiters;
            return NextToken();
        }

        /// <summary>
        /// Removes all empty strings from the token list
        /// </summary>
        private void RemoveEmptyStrings()
        {
            for (int index=0; index < this.elements.Count; index++)
                if ((string)this.elements[index]== "")
                {
                    this.elements.RemoveAt(index);
                    index--;
                }
        }
    }

    /// <summary>
    /// Summary description for Class1.
    /// </summary>
	class PD4Test
	{
		static char quote = '\"';
		static String USAGE = "Usage: pd4net "+quote+"<url>"+quote+" <htmlWidth> [pageFormatName|WxH] " +
			"[-pdfa] [-smarttablesplit] [-permissions <NUMBER>] [-bookmarks <HEADINGS|ANCHORS>] " +
			"[-orientation <PORTRAIT|LANDSCAPE>] [-insets <T,L,B,R,><mm|pt>] [-bgcolor <#RGB>] [-bgimage "+quote+"<url>"+quote+"] [-ttf <ttf_fonts_dir>]" + 
			"[-addstyle <CSS code>] [-pdfforms] [-protectpud] [-adjustwidth] [-fitapage] [-nohyperlinks] [-noimagesplit] " +
			"[-author <author name>] [-title <title override>] [-cookie <name> <value>] [-param <name> <value>] [-header "+quote+"<header HTML code>"+quote+"] " +
			"[-footer "+quote+"<footer HTML code>"+quote+"] [-pagerange <page>] [-out <output_file_path>] [-password <password>] [-debug]" +
			"-type (PDF|RTF|RTF_WMF|Image|PNG8|PNG24|TIFF|TIFFCOMP)";

		private static String in_S;
		private static String type = "PDF";
		private static int width = -1;
		private static String format = "A4";
		private static int permissions = -1;
		private static String bookmarks;
		private static String orientation;
		private static String insets;
		private static String ttf;
		private static String out_S;
		private static bool smart = false;

		private static bool forms = false;
		private static bool physUnits = false;
		private static bool adjustWidth = false;
		private static bool fitPage = false;
		private static bool hyperlinks = true;
		private static bool imgSplit = true;
		private static bool pdfa = false;
		private static bool debug = false;
	
		private static String author;
		private static String title;
		private static String range;
		private static String password;

		private static String bgcolor;
		private static String bgimage;

		private static ArrayList styles;
	
		private static Hashtable cookies;
		private static Hashtable dynParams;
	
		private static String header;
		private static String footer;
		private static int header_h = -1;
		private static int footer_h = -1;

		private static bool consumeCheck(String argv, String key) 
		{
			argv = argv.ToLower();
			key = key.ToLower();
			if(!key.StartsWith("-")) key = "-" + key;
			return argv.Equals(key);
		}

		private static bool consume(String argv, String key, ref bool v) 
		{
			if (consumeCheck(argv, key)) 
			{
				v = true;
				return true;
			}
			return false;
		}

		private static bool consume(String[] args, ref int i, String key, ref int v, String msg) 
		{
			if (consumeCheck(args[i], key)) 
			{
				++i;
				if ( i == args.Length ) 
					wrongParam(msg);
				else
					v = Int32.Parse(args[i]);
				return true; 
			}
			return false;
		}

		private static bool consume(String[] args, ref int i, String key, ref String v, String msg) 
		{
			if (consumeCheck(args[i], key)) 
			{
				++i;
				if ( i == args.Length ) 
					wrongParam(msg);
				else
					v = args[i];
				return true;
			}
			return false;
		}

		private static bool consume(String[] args, ref int i, String key, ref String v, String msg, String[] accepted) 
		{
			if (consumeCheck(args[i], key)) 
			{
				++i;
				if ( i == args.Length ) 
					wrongParam(msg);
				else 
				{
					String vv = args[i].ToLower();
					for(int j =0; j < accepted.Length; ++j) 
						if(accepted[j].ToLower().Equals(vv)) 
						{
							v = args[i];
							return true;
						}
					wrongParam(msg);
				}
				return true;
			}
			return false;
		}

		private static bool consume(String[] args, ref int i, String key, ref ArrayList v, String msg) 
		{
			if (consumeCheck(args[i], key)) 
			{
				++i;
				if (v == null) v = new ArrayList();
				if ( i == args.Length ) 
					wrongParam(msg);
				else if (v.IndexOf(args[i]) < 0)
					v.Add(args[i]);
				return true; 
			}
			return false;
		}

		private static bool consume(String[] args, ref int i, String key, ref Hashtable v, String msg) 
		{
			if (consumeCheck(args[i], key)) 
			{
				++i;
				if (v == null) v = new Hashtable();
				if ( i + 1 == args.Length ) 
					wrongParam(msg);
				else{ 
					String k = args[i];
					String vv = args[i + 1];
					++i;
					v.Add(k, vv);
				}
				return true; 
			}
			return false;
		}

		private static void wrongParam(string msg) 
		{
			Console.Out.WriteLine("invalid parameter: " + msg);
			Console.Out.WriteLine( USAGE );
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			bool formatSet = false;
			FileInfo ff;
			String[] knownTypes = {
				"PDF", "RTF", "Image", "RTF_WMF",
				PD4Constants.PDF, PD4Constants.PDFA,
				PD4Constants.RTF, PD4Constants.RTF_WMF,
				PD4Constants.PNG24, PD4Constants.PNG8, PD4Constants.TIFF, PD4Constants.TIFF_COMPRESSED
			};
			String[] knownMarks = {"HEADINGS", "ANCHORS"};
			String[] knownOrientation = {"PORTRAIT", "LANDSCAPE"};
            
			for( int i = 0; i < args.Length; i++ ) 
			{
				if (consume(args[i], "-pdfforms", ref forms)) continue;
				if (consume(args[i], "-protectpud", ref physUnits)) continue;
				if (consume(args[i], "-adjustwidth", ref adjustWidth)) continue;
				if (consume(args[i], "-fitapage", ref fitPage)) continue;
				if (consume(args[i], "-nohyperlinks", ref hyperlinks)) continue;
				if (consume(args[i], "-pdfa", ref pdfa)) continue;
				if (consume(args[i], "-noimagesplit", ref imgSplit)) continue;
				if (consume(args[i], "-debug", ref debug)) continue;
				if (consume(args[i], "-smarttablesplit", ref smart)) continue;

				if (consume(args, ref i, "-type", ref type, "missing type, one of (PDF, RTF|RTF_WMF, Image|PNG8|PNG24|TIFF)", knownTypes)) continue;

				if (consume(args, ref i, "-bgcolor", ref bgcolor, "invalid parameter: color is missing (Example: -bgcolor "+quote+"#ffffcc"+quote+" or  -bgcolor 0xffffcc)")) continue;
				if (consume(args, ref i, "-bgimage", ref bgimage, "invalid parameter: background URL is missing (Example: -bgimage "+quote+"http://pd4ml.com/i/blank.jpg"+quote+" or -bgimage "+quote+"file:/resources/images/blank.jpg"+quote+")")) continue;
				if (consume(args, ref i, "-author", ref author, "invalid parameter: author name is missing")) continue;
				if (consume(args, ref i, "-title", ref title, "invalid parameter: title override string is missing")) continue;
				if (consume(args, ref i, "-password", ref password, "invalid parameter: password string is missing")) continue;
				if (consume(args, ref i, "-pagerange", ref range, "invalid parameter: no output page range definition string (Example: -pagerange "+quote+"2-3,7+"+quote+")")) continue;
				if (consume(args, ref i, "-header", ref header, "invalid parameter: PDF header definition HTML code is missing (Example: -header "+quote+"<div width=100% align=right>$[page] of $[total]</div>"+quote+")")) continue;
				if (consume(args, ref i, "-footer", ref footer, "invalid parameter: PDF footer definition HTML code is missing (Example: -footer "+quote+"<div width=100% align=right>$[page] of $[total]</div>"+quote+")")) continue;
				if (consume(args, ref i, "-header-height", ref header_h, "invalid parameter: wrong height spec (Example: -header-height 40)")) continue;
				if (consume(args, ref i, "-footer-height", ref footer_h, "invalid parameter: wrong height spec (Example: -footer-height 40)")) continue;

				if (consume(args, ref i, "-addstyle", ref styles, "invalid parameter: style definition is missing  (Example: -addstyle "+quote+"IMG, TR {page-break-inside: avoid}"+quote+")")) continue;
				if (consume(args, ref i, "-cookie", ref cookies, "invalid parameter: incomplete cookie data  (Example: -cookie JSESSIONID "+quote+"9034657927465;path=/"+quote+")")) continue;
				if (consume(args, ref i, "-param", ref dynParams, "invalid parameter: incomplete cookie data  (Example: -param pd4ml.basic.authentication "+quote+"user:password"+quote+")")) continue;

				if (consume(args, ref i, "-permissions", ref permissions, "missing permissions number (a sum of single permission codes)")) continue;
				if (consume(args, ref i, "-bookmarks", ref bookmarks, "invalid parameter: missing bookmark type (HEADINGS or ANCHORS)", knownMarks)) continue;
				if (consume(args, ref i, "-orientation", ref orientation, "invalid parameter: missing orientation type (PORTRAIT or LANDSCAPE)", knownOrientation)) continue;
				if (consume(args, ref i, "-insets", ref insets, "invalid parameter: missing insets (for exampe: -insets 5,10,5,5mm)")) continue;

				if (consume(args, ref i, "-out", ref out_S, "invalid parameter: missing output file name")) continue;
				if (consume(args, ref i, "-ttf", ref ttf, "invalid parameter: missing TTF fonts directory"))
				{ 
					if(ttf!= null) 
					{ 
						ff = new FileInfo(ttf);
						ttf = ff.FullName;
					}
					continue;
				}

				if ( args[i].StartsWith("-")  ) 
				{
					wrongParam("unknown: " + args[i]);
					return; 
				}

				if ( in_S == null ) 
				{
					string ss = args[i];
					if(!ss.StartsWith("http://") && !ss.StartsWith("https://") && !ss.StartsWith("file://")) 
					{
						ff = new FileInfo(ss);
						in_S = "file:///" + ff.FullName;
					}
					else
						in_S = ss;
					continue;
				}

				if ( width == -1 ) 
				{
					width = Int32.Parse(args[i]);
					continue;
				}

				if ( !formatSet ) 
				{
					formatSet = true;
					format = args[i];
					continue;
				}

				wrongParam("unexpected parameter: "  + args[i]);
				return; 
			}
            
			if ( in_S == null ) 
			{
				wrongParam("source URL is missing");
				return;
			}

			if ( width == -1 ) 
			{
				wrongParam("HTML width parameter is missing");
				return;
			}

			PD4Test converter = new PD4Test();

			if(type == null || type.Length == 0)
				type = "PDF";
			type = type.Trim().ToLower();

			switch(type) 
			{
				case "pdf":
					type = pdfa ? PD4Constants.PDFA : PD4Constants.PDF;
					break;
				case "rtf":
					type = PD4Constants.RTF;
					break;
				case "rtf_wmf":
					type = PD4Constants.RTF_WMF;
					break;
				case "image":
					type = PD4Constants.PNG24;
					break;
				case PD4Constants.PNG8:
				case PD4Constants.PNG24:
				case PD4Constants.TIFF:
				case PD4Constants.TIFF_COMPRESSED:
					break;
				case PD4Constants.PDFA:
				case PD4Constants.RTF_WMF:
					break;
				default:
					wrongParam("wrong type: " + type);
					return;
			}
			
			converter.generate(type); 
		}

		private bool coverInsets(String insets, ref Rectangle rr, ref char uom)
		{
			if(insets == null) return false;
			Tokenizer st = new Tokenizer( insets, ",");
			rr = new Rectangle(0,0,0,0);
			uom = 'P';
			try 
			{
				int top = Int32.Parse(st.NextToken());
				int left = Int32.Parse(st.NextToken());
				int bottom = Int32.Parse(st.NextToken()) + top;
				String rightStr = st.NextToken();
				String units = null;
				int right = 0;
                    
				if(st.Count > 4) 
				{
					right = Int32.Parse(rightStr) + left;
					units = st.NextToken();
				}
				else 
				{ 
					units = rightStr.Substring(rightStr.Length - 2, 2);
					rightStr = rightStr.Substring(0, rightStr.Length -2);
					right = Int32.Parse(rightStr) + left;
				}

				rr.X = left;
				rr.Y = top;
				rr.Width = right - left;
				rr.Height = bottom - top;
				if ("MM".Equals(units.ToUpper())) 
				{
					uom = 'M';
				} 
				else 
				{
					uom = 'P';
				}
			} 
			catch (Exception) 
			{
				rr = new Rectangle(0,0,0,0);
				uom = 'P';
				throw new Exception(
					"Invalid page insets (top, left, bottom, right, units): "
					+ insets );
			}
			return true;
		}
        
		private void generate(String type)
		{
			PD4ML pd4ml = new PD4ML();

			// insets
			Rectangle ins = new Rectangle(0,0,0,0);
			char uom = 'P';
			if(coverInsets(insets, ref ins, ref uom)) 
			{
				if (uom == 'M') 
				{
					pd4ml.PageInsetsMM = ins;
				} 
				else 
				{
					pd4ml.PageInsets = ins;
				}
			}
            
			// sizes
			if(orientation == null) orientation = "portrait";
			System.Drawing.Size pageSize = new Size(0, 0);
			Tokenizer st = new Tokenizer(format, "x");
			if(st.Count == 2) 
			{
				pageSize.Width = Int32.Parse(st.NextToken());
				pageSize.Height = Int32.Parse(st.NextToken());
			}
			else
				pageSize = PD4Constants.getSizeByName(format);

			if (pageSize.Width < 1 || pageSize.Height < 1) 
			{
				wrongParam("wrong page size");
				return;
			}

			if ("PORTRAIT".Equals(orientation.ToUpper())) 
				pd4ml.PageSize = pageSize;
			else 
				pd4ml.PageSize = pd4ml.changePageOrientation( pageSize ); 

			// options
			pd4ml.HtmlWidth = width;
			pd4ml.enableTableBreaks(smart);
			if ( ttf != null && ttf.Length > 0 ) pd4ml.useTTF( ttf, true );
			if (forms) pd4ml.generatePdfForms(true, "Arial");
			if (physUnits) pd4ml.protectPhysicalUnitDimensions();
			if (adjustWidth) pd4ml.adjustHtmlWidth();
			if (fitPage) pd4ml.fitPageVertically();
			if (!hyperlinks) pd4ml.disableHyperlinks();
			if (!imgSplit) pd4ml.enableImgSplit(false);
			if (author != null ) pd4ml.AuthorName = author;
			if (title != null ) pd4ml.DocumentTitle = title;
			if (range != null ) pd4ml.OutputRange = range;
			if (dynParams != null ) pd4ml.DynamicParams = dynParams;
			if (debug) pd4ml.enableDebugInfo();
			pd4ml.OutputFormat = type;
			if (pdfa) pd4ml.generatePdfa(true);

			// permissions
            
			if (permissions != -1) 
			{
				if ( password != null ) 
					pd4ml.setPermissions(password, permissions, true);
				else 
					pd4ml.setPermissions("empty", permissions, true);
			} 
			else 
			{
				if ( password != null ) 
					pd4ml.setPermissions(password, -4 /*0xfffffffc*/, true);
			}

			// bookmarks
			if ( bookmarks != null ) 
			{
				if ( "ANCHORS".Equals(bookmarks.ToUpper()) )
					pd4ml.generateOutlines(false);
				else if ( "HEADINGS".Equals(bookmarks.ToUpper()) ) 
					pd4ml.generateOutlines(true);
			}

			// styles and cookies
			if (styles != null ) 
				for(int j = 0; j < styles.Count; ++j)
					pd4ml.addStyle((String)styles[j], true);

			if (cookies != null ) 
			{
				IEnumerator ii = cookies.Keys.GetEnumerator();
				while (ii.MoveNext()) 
				{
					String key = (String)ii.Current;
					String val = (String)cookies[key];
					pd4ml.setCookie(key, val);
				}
			}

			// headers and footers
			if ( footer != null && footer.Trim().Length > 0 || 
				bgimage != null && bgimage.Trim().Length > 0 || 
				bgcolor != null && bgcolor.Trim().Length > 0  ) 
			{
				PD4PageMark footerObj = new PD4PageMark();
				footerObj.AreaHeight = footer_h; // autocompute is -1
				try 
				{
					if ( footer != null && footer.Trim().Length > 0 ) 
						footerObj.setHtmlTemplate(footer);
					if ( bgimage != null && bgimage.Trim().Length > 0 ) 
						footerObj.setPageBackgroundImageUrl(bgimage);
					if ( bgcolor != null && bgcolor.Trim().Length > 0 ) 
					{
						bgcolor = bgcolor.Trim().Replace("#", "0x");
						Color c = Color.FromArgb(Int32.Parse(bgcolor, System.Globalization.NumberStyles.HexNumber));
						footerObj.setPageBackgroundColor(c);
					}
					pd4ml.PageFooter = footerObj;
				} 
				catch (Exception e) 
				{
					wrongParam("wrong footer qualifier, check background image/color also. " + e.StackTrace);
				}
			}

			if ( header != null && header.Trim().Length > 0) 
			{
				PD4PageMark headerObj = new PD4PageMark();
				headerObj.setHtmlTemplate(header);

				if ( bgimage != null && bgimage.Trim().Length > 0 ) 
					headerObj.setPageBackgroundImageUrl(bgimage);
				if ( bgcolor != null && bgcolor.Trim().Length > 0 ) 
				{
					bgcolor = bgcolor.Trim().Replace("#", "0x");
					Color c = Color.FromArgb(Int32.Parse(bgcolor, System.Globalization.NumberStyles.HexNumber));
					headerObj.setPageBackgroundColor(c);
				}
				headerObj.AreaHeight = header_h; // autocompute is -1
				pd4ml.PageHeader = headerObj;
			}

			/*
			pd4ml.setPermissions("", 
				PD4Constants.AllowPrint 
				| PD4Constants.AllowCopy 
				| PD4Constants.AllowFillingForms
				| PD4Constants.AllowAnnotate
				| PD4Constants.AllowModify
				| PD4Constants.AllowDegradedPrint
				| PD4Constants.AllowContentExtraction, 
				true);
				*/

			/*
			PD4PageMark hh = new PD4PageMark();
			hh.AreaHeight = 40;
			hh.Color = Color.FromArgb(0, 0, 0);
			hh.setFont(new Font("Times New Roman", 12.0f, FontStyle.Italic));
			hh.setPageNumberTemplate("header: $[page] of $[total]");
			hh.PageNumberAlignment = PD4PageMark.CENTER_ALIGN;
			hh.setPageBackgroundColor(Color.FromArgb(0x99ccff));
			
			pd4ml.PageHeader = hh;
			pd4ml.PageInsets = new Rectangle(1, 100, 1, 1);
			*/

			// output file
			if (out_S == null || out_S.Equals("-"))
			{ 
				pd4ml.render(new Uri(in_S), Console.Out);
			}
			else if (out_S.Equals("-SIZE"))
			{ 
				MemoryStream ms = new MemoryStream();
				pd4ml.render(new Uri(in_S), ms);
				byte[] bb = ms.ToArray();
				Console.Out.WriteLine("Bytes of PDF = " + bb.Length);
				ms.Close();
			}
			else
				pd4ml.render(
					new Uri(in_S), 
					new System.IO.FileStream(out_S, System.IO.FileMode.Create));
		}

		private void encodeImage(Bitmap bmp, Stream fs)
		{
			bmp.Save(fs, System.Drawing.Imaging.ImageFormat.Png);
		}
	}
}
